#pragma GCC optimize ("-Os")
#pragma GCC push_options

#include <avr/interrupt.h>
#include <avr/io.h>
#include <avr/wdt.h>
#include "Arduino.h"
#include <EEPROM.h>
#include <LiquidCrystal_I2C.h>
#include <Wire.h>
#include <avr/pgmspace.h>
#include "Setting.h"
#include "Variables.h"
#include "Encoder.h"
#include "Button.h"
#include "PID.h"

LiquidCrystal_I2C lcd(0x27, 16, 2);
modedEncoder *encoder;
SWITCH toggleSwitch(A0);
modedPID_Tuner tuner1;
modedPID_Tuner tuner2;
modedPID regulator1(TUNE_P_PID, TUNE_I_PID, TUNE_D_PID, SENSOR_SAMPLING_TIME);
modedPID regulator2(TUNE_P_PID, TUNE_I_PID, TUNE_D_PID, SENSOR_SAMPLING_TIME);

float readTemp_Max6675(byte p) {
  int a = 0;
  if (p == 1) {
    PORTB &= B11111110;
  } else if (p == 2) {
    PORTB &= B11110111;
  }
  __asm__ __volatile__ ("nop\n\t");
  for (int8_t i = 15; i >= 0; i--) {
    PORTB |= B00100000;
    __asm__ __volatile__ ("nop\n\t");
    if (((PINB & B00010000) >> 4)) {
      a |= (1 << i);
    }
    PORTB &= B11011111;
    __asm__ __volatile__ ("nop\n\t");
  }
  if (p == 1) {
    PORTB |= B00000001;
  }
  else if (p == 2) {
    PORTB |= B00001000;
  }
  if (a & 0x4) {
    return 255;
  }
  else return ((a >> 3) * 0.25);
}

float getTempHuman(byte sel) {
  float t;
  float tem = 0.0;
  t = readTemp_Max6675(sel) * (0.6 + 0.4);
  for (int i = 0; i < 9; i++) {
    tem += t;
  }
  return tem / 9;
}

double EEPROM_readDouble(int address) {
  double value = 0.0;
  byte* p = (byte*)(void*)&value;
  for (int i = 0; i < sizeof(value); i++) {
    *p++ = EEPROM.read(address++);
  }
  return value;
}

void beep_(uint16_t dur, uint16_t frq) {
  uint16_t per = 500000 / frq;
  dur = dur / (per / 250) * 2;
  for (uint16_t i = 0; i < dur; i++) {
    PORTD |= B10000000;
    delay_us(per);
    PORTD &= B01111111;
    delay_us(per);
  }
}

void delay_us(uint16_t tic_us) {
  tic_us *= 4;
  __asm__ volatile("1: sbiw %0,1"
                   "\n\t""brne 1b":
                   "=w" (tic_us):
                   "0" (tic_us));
}

void saveDefPID() {
  EEPROM_writeDouble(KpU_Prof_Address,
                     pgm_read_float(&DF_PID_P[1]));
  EEPROM_writeDouble(KiU_Prof_Address,
                     pgm_read_float(&DF_PID_I[1]));
  EEPROM_writeDouble(KdU_Prof_Address,
                     pgm_read_float(&DF_PID_D[1]));
  EEPROM_writeDouble(KpD_Prof_Address,
                     pgm_read_float(&DF_PID_P[0]));
  EEPROM_writeDouble(KiD_Prof_Address,
                     pgm_read_float(&DF_PID_I[0]));
  EEPROM_writeDouble(KdD_Prof_Address,
                     pgm_read_float(&DF_PID_D[0]));
  return;
}

void loadPIDParam() {
  KpU = EEPROM_readDouble(KpU_Prof_Address);
  KiU = EEPROM_readDouble(KiU_Prof_Address);
  KdU = EEPROM_readDouble(KdU_Prof_Address);

  KpD = EEPROM_readDouble(KpD_Prof_Address);
  KiD = EEPROM_readDouble(KiD_Prof_Address);
  KdD = EEPROM_readDouble(KdD_Prof_Address);
  return;
}

void saveDefProfile() {
  for (int thisCount = 0; thisCount < 9; thisCount++) {
    EEPROM.update(thisCount * 31,
                  (pgm_read_byte(&def_profileSteps[thisCount])));
  }
  for (int thisCount = 0; thisCount < 9; thisCount++) {
    EEPROM.update(thisCount * 31 + 1,
                  (pgm_read_byte(&def_Setpoint2[thisCount])));
  }

  for (int profCount = 0; profCount < 9; profCount++) {
    for (int thisCount = 0; thisCount < 9; thisCount++) {
      EEPROM.update(profCount * 31 + thisCount + 2,
                    (pgm_read_byte(&def_temperatureStep[profCount][thisCount])));
    }
  }

  for (int profCount = 0; profCount < 9; profCount++) {
    for (int thisCount = 0; thisCount < 9; thisCount++) {
      EEPROM.update(profCount * 31 + thisCount + 11,
                    (pgm_read_byte(&def_rampRateStep[profCount][thisCount])));
    }
  }

  for (int profCount = 0; profCount < 9; profCount++) {
    for (int thisCount = 0; thisCount < 9; thisCount++) {
      EEPROM.update(profCount * 31 + thisCount + 20,
                    (pgm_read_byte(&def_dwellTimerStep[profCount][thisCount])));
    }
  }

  for (int thisCount = 0; thisCount < 9; thisCount++) {
    EEPROM.update(thisCount * 31 + 29,
                  (pgm_read_byte(&def_PWR_D[thisCount])));
  }

  for (int thisCount = 0; thisCount < 9; thisCount++) {
    EEPROM.update(thisCount * 31 + 30,
                  (pgm_read_byte(&def_PWR_U[thisCount])));
  }
  return;
}

void loadProfile() {
  profile.profileSteps = EEPROM.read((currentProfile - 1) * 31);
  profile.Setpoint2 = EEPROM.read((currentProfile - 1) * 31 + 1);
  for (int i = 0; i < 9; i + 1) {
    profile.temperatureStep[i] = EEPROM.read((currentProfile - 1) * 31 + i + 2);
    i++;
  }
  for (int i = 0; i < 9; i + 1) {
    profile.rampRateStep[i] = EEPROM.read((currentProfile - 1) * 31 + i + 11);
    i++;
  }
  for (int i = 0; i < 9; i + 1) {
    profile.dwellTimerStep[i] = EEPROM.read((currentProfile - 1) * 31 + i + 20);
    i++;
  }
  profile.pwr_BOTTOM = EEPROM.read((currentProfile - 1) * 31 + 29);
  profile.pwr_TOP = EEPROM.read((currentProfile - 1) * 31 + 30);
  return;
}

void initEeprom() {
  byte firstTime = EEPROM.read(eeprom_id_address);
  if (firstTime == eeprom_id) {
    currentProfile = 1;
    loadProfile();
    loadPIDParam();
    return;
  } else {
    lcd.clear();
    lcd.setCursor(0, 0);
    lcd.print(F("ERASE"));
    lcd.setCursor(0, 1);
    for (int i = 0; i < 15; i++) {
      lcd.print(F(".")); delay(100);
    }
    for (int i = 0; i < EEPROM.length(); i++) {
      EEPROM.write(i, 0);
    }
    EEPROM.update (500, eeprom_id);

    lcd.clear();
    lcd.setCursor(0, 0);
    lcd.print(F("RESTORE"));
    saveDefProfile();
    saveDefPID();
    lcd.setCursor(0, 1);
    for (int i = 0; i < 15; i++) {
      lcd.print(F("."));
      delay(150);
    }
    delay(1000);
    currentProfile = 1;
    loadProfile();
    loadPIDParam();
    initEepromComplete = true;
  }
  return;
}

void sReset() {
  asm volatile ("jmp 0");
}

ISR(TIMER1_OVF_vect) {
  encoder->service();
}

ISR(TIMER2_COMPA_vect) {
  PORTB |= B00000010;
}

ISR(TIMER2_COMPB_vect) {
  PORTB |= B00000100;
}

void TopHeaterON(int dim) {
  OCR2B = 151 - dim;
}

void TopHeaterOFF(void) {
  OCR2B = 150;
}

void BottomHeaterON(int dim) {
  OCR2A = 151 - dim;
}

void BottomHeaterOFF(void) {
  OCR2A = 150;
}

void Dimming() {
  PORTB &= B11111101;
  PORTB &= B11111011;
  TCNT2 = 0;
  if (Status == 1) {
    sprintf (buf, "OK%03d%03d%03d%03d%03d\r\n",
             P_U_Serial,
             P_D_Serial,
             tc1,
             tc2,
             currentProfile);
  }
  if (Status == 2) {
    byte pdMap = map(P_D, 0, 150, 0, 100);
    sprintf (buf, "OK%03d%03d%03d%03d%03d\r\n",
             tunerAccu2,
             pdMap,
             tc1,
             tc2,
             Status);
  }
  if (Status == 4) {
    byte puMap = map(P_U, 0, 150, 0, 100);
    sprintf (buf, "OK%03d%03d%03d%03d%03d\r\n",
             tunerAccu1,
             puMap,
             tc1,
             tc2,
             Status);
  }
}

void EEPROM_writeDouble(int address, double value) {
  byte* p = (byte*)(void*)&value;
  for (volatile int i = 0; i < sizeof(value); i++) {
    EEPROM.update(address++, *p++);
    delay(25);
  }
}

void setup() {
  Serial.begin(9600);

  DDRD &= B11111011;

  DDRD &= B11110111;
  PORTD |= B00001000;

  DDRD |= B00010000;
  PORTD &= B11101111;

  DDRD &= B11011111;

  DDRD &= B10111111;

  DDRD |= B10000000;
  PORTD &= B01111111;

  DDRB |= B00000001;
  PORTB |= B00000001;

  DDRB |= B00000010;

  DDRB |= B00000100;

  DDRB |= B10001000;
  PORTB |= B00001000;

  DDRB &= B11101111;

  DDRB |= B00100000;

  DDRC &= B11111110;

  DDRC |= B00000010;
  PORTC &= B11111101;                          //protection Relay LOW

  cli();

  TCCR1B = _BV(WGM13);
  TCCR1A = 0;

  TCCR2A = 0;
  TCCR2B = 0;

  ICR1 =   (F_CPU / 2000000) * 10000;
  TCCR1B = _BV(WGM13) | _BV(CS10);

  OCR2A = 152;
  OCR2B = 152;

  TCCR2B |= (1 << WGM22);
  TCCR2B |= (1 << CS20);
  TCCR2B |= (1 << CS21);
  TCCR2B |= (1 << CS22);

  TIMSK1 = _BV(TOIE1);

  TIMSK2 |= (1 << OCIE2A);
  TIMSK2 |= (1 << OCIE2B);
  sei();

  encoder = new modedEncoder(6, 5, 2);

  attachInterrupt(1, Dimming, RISING);

  toggleSwitch.init(500, 3000);

  Wire.begin();
  Wire.setClock(10000);
  lcd.begin();
  lcd.backlight();
  lcd.createChar(0, GC);
  lcd.createChar(1, Th);
  lcd.createChar(2, Blh);
  lcd.createChar(3, Bch);
  lcd.createChar(4, Brh);
  lcd.createChar(5, customChar);
  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print(F("-=SouthBridge"));
  lcd.write(byte(5));
  lcd.print(F("=-"));
  lcd.setCursor(2, 1);
  lcd.print(F("BGAIRS v"));
  lcd.print(VERSION);
  lcd.print(F("."));
  lcd.print(REVISION);

  beep_(200, 523);
  beep_(200, 659);
  beep_(200, 784);
  beep_(200, 1046);
  delay(3000);

  initEeprom();
  lcd.clear();

  nextRead1 = millis();
}

void loop() {
labelbegin:
  unsigned long currentMillis = millis();

  toggleSwitchState = toggleSwitch.status();
  toggleSwitchState = !toggleSwitchState;

  value += encoder->getValue();
  modedEncoder::Button b = encoder->getButton();
  increment = (value - last);
  decrement = (last - value);

  if (screen == 14 || alarmOn) {
    if (alarmCount < 15 && toggleSwitchState) {
      alarmOn = true;
      beep_(40, 1046); delay(80);
      beep_(40, 1046); delay(200);
      beep_(40, 1046); delay(80);
      beep_(40, 1046); delay(200);
      alarmCount++;
    } else {
      alarmCount = 0;
      alarmOn = false;
    }
  }

  if (millis() > nextRead1) {
    nextRead1 = millis() + SENSOR_SAMPLING_TIME;
    Input1 = getTempHuman(1);
    Input2 = getTempHuman(2);
  }

  switch (screen) {
    case 0:

      TopStart = false;
      currentStep = 1;
      counter = 0;
      counterB = 0;
      setpointRamp = 0;
      flag = false;
      flagB = false;

      showIdleScreen();
      lcdTemp();

      if (toggleSwitchState) {
        beep_(200, 1046);
        delay(100);
        beep_(40, 1094);
        curCount = 0;
        P_D = 0;
        P_U = 0;
        updateScreen = true;
        startTemp = 0;
        startTempB = 0;
        PORTC |= B00000010;                                     //protection Relay HIGH
        nextRead2 = millis();
        nextRead3 = millis();

        byte pwr1 = profile.pwr_TOP;
        byte pwrMaxOut1;
        pwrMaxOut1 = map(pwr1, 0, 100, 0, 150);
        regulator1.setDirection(0);
        regulator1.setDt(SENSOR_SAMPLING_TIME);
        regulator1.setLimits(0, pwrMaxOut1);

        byte pwr2 = profile.pwr_BOTTOM;
        byte pwrMaxOut2;
        pwrMaxOut2 = map(pwr2, 0, 100, 0, 150);
        regulator2.setDirection(0);
        regulator1.setDt(SENSOR_SAMPLING_TIME);
        regulator2.setLimits(0, pwrMaxOut2);

        screen = 11;
        Status = 1;
        startSerialCom();
      }

      if (value > last) { //encoder plus
        last = value;
        nex = true;
        if (nex && ( millis() - ms_button) > 250) {
          ms_button =  millis();
          beep_(80, 1046);
          nex = false;
          currentProfile++;
          if (currentProfile >= 9) {
            currentProfile = 9;
          }
        }
      }

      if (value < last) { //encoder minus
        last = value;
        prev = true;
        if (prev && ( millis() - ms_button) > 250) {
          ms_button =  millis();
          beep_(80, 1046);
          prev = false;
          currentProfile--;
          if (currentProfile <= 1) {
            currentProfile = 1;
          }
        }
      }

      if (b != modedEncoder::Open) { //encoder button
        switch (b) {
          case modedEncoder::Clicked:
            beep_(80, 1046);
            screen = 17;
            selection2 = 0;
            updateScreen = true;
            break;
          case modedEncoder::Released:
            beep_(120, 1046);
            Status = 3;
            screen = 15;
            nextRead2 = millis();
            updateScreen = true;
            break;
          case modedEncoder::DoubleClicked:
            beep_(80, 1046);
            updateScreen = true;
            screen = 8;
            pidCursor = 0;
            break;
        }
      }
      break;

    case 1:
      showProfEditStepScr();

      if (value > last) { //encoder plus
        last = value;
        nex = true;
        if (nex && ( millis() - ms_button) > 250) {
          ms_button =  millis();
          beep_(80, 1046);
          nex = false;
          profile.profileSteps++;
          if (profile.profileSteps >= 9) {
            profile.profileSteps = 9;
          }
        }
      }

      if (value < last) { //encoder minus
        last = value;
        prev = true;
        if (prev && ( millis() - ms_button) > 250) {
          ms_button =  millis();
          beep_(80, 1046);
          prev = false;
          profile.profileSteps--;
          if (profile.profileSteps <= 1) {
            profile.profileSteps = 1;
          }
        }
      }

      if (b != modedEncoder::Open) { //encoder button
        switch (b) {
          case modedEncoder::Clicked:
            beep_(80, 1046);
            saveProfileSteps();
            screen = 17;
            updateScreen = true;
            break;
        }
      }
      break;

    case 2:
      showProfSP2Edit();
      if (value > last) { //encoder plus
        last = value;
        beep_(40, 1046);
        profile.Setpoint2 += increment;
        if (profile.Setpoint2 >= 230) {
          profile.Setpoint2 = 230;
        }
      }

      if (value < last) { //encoder minus
        last = value;
        beep_(40, 1046);
        profile.Setpoint2 -= decrement;
        if (profile.Setpoint2 <= 80) {
          profile.Setpoint2 = 80;
        }
      }

      if (b != modedEncoder::Open) { //encoder button
        switch (b) {
          case modedEncoder::Clicked:
            beep_(80, 1046);
            saveProfileStpoint2();
            screen = 17;
            editStep = 0;
            updateScreen = true;
            break;
        }
      }
      break;

    case 3:
      showTempStpEditScr();
      if (value > last) { //encoder plus
        last = value;
        beep_(40, 1046);
        profile.temperatureStep[editStep] += increment;
        if (profile.temperatureStep[editStep] >= 238) {
          profile.temperatureStep[editStep] = 238;
        }
      }

      if (value < last) { //encoder minus
        last = value;
        beep_(40, 1046);
        profile.temperatureStep[editStep] -= decrement;
        if (profile.temperatureStep[editStep] <= 80) {
          profile.temperatureStep[editStep] = 80;
        }
      }

      if (b != modedEncoder::Open) { //encoder button
        switch (b) {
          case modedEncoder::Clicked:
            beep_(80, 1046);
            if (editStep + 1 == profile.profileSteps) {
              saveProfileTempStep();
              editStep = 0;
              screen = 17;
              updateScreen = true;
            } else {
              saveProfileTempStep();
              editStep++;
              updateScreen = true;
            }
            break;
          case modedEncoder::Released:
            beep_(120, 1046);
            saveProfileTempStep();
            loadProfile();
            editStep = 0;
            screen = 17;
            updateScreen = true;
            break;
        }
      }
      break;

    case 4:
      showEditStepScreen();
      if (value > last) { //encoder plus
        last = value;
        nex = true;
        if (nex && ( millis() - ms_button) > 250) {
          ms_button =  millis();
          beep_(80, 1046);
          nex = false;
          profile.rampRateStep[editStep]++;
          if (profile.rampRateStep[editStep] >= 30) {
            profile.rampRateStep[editStep] = 30;
          }
        }
      }

      if (value < last) { //encoder minus
        last = value;
        prev = true;
        if (prev && ( millis() - ms_button) > 250) {
          ms_button =  millis();
          beep_(80, 1046);
          prev = false;
          profile.rampRateStep[editStep]--;
          if (profile.rampRateStep[editStep] <= 1) {
            profile.rampRateStep[editStep] = 1;
          }
        }
      }

      if (b != modedEncoder::Open) { //encoder button
        switch (b) {
          case modedEncoder::Clicked:
            beep_(80, 1046);
            if (editStep + 1 == profile.profileSteps) {
              saveProfileRampStep();
              updateScreen = true;
              editStep = 0;
              screen = 17;
            } else {
              saveProfileRampStep();
              editStep++;
              updateScreen = true;
            }
            break;
          case modedEncoder::Released:
            beep_(120, 1046);
            saveProfileRampStep();
            loadProfile();
            screen = 17;
            editStep = 0;
            updateScreen = true;
            break;
        }
      }
      break;

    case 5:
      showProfEditDwlScr();
      if (value > last) { //encoder plus
        last = value;
        beep_(40, 1046);
        profile.dwellTimerStep[editStep] += increment;
        if (profile.dwellTimerStep[editStep] >= 240) {
          profile.dwellTimerStep[editStep] = 240;
        }
      }

      if (value < last) { //encoder minus
        last = value;
        beep_(40, 1046);
        profile.dwellTimerStep[editStep] -= decrement;
        if (profile.dwellTimerStep[editStep] <= 1) {
          profile.dwellTimerStep[editStep] = 1;
        }
      }

      if (b != modedEncoder::Open) { //encoder button
        switch (b) {
          case modedEncoder::Clicked:
            beep_(80, 1046);
            if (editStep + 1 == profile.profileSteps) {
              saveProfileDwell();
              editStep = 0;
              screen = 17;
              updateScreen = true;
            } else {
              saveProfileDwell();
              editStep++;
              updateScreen = true;
            }
            break;
          case modedEncoder::Released:
            beep_(120, 1046);
            saveProfileDwell();
            loadProfile();
            delay(200);
            screen = 17;
            editStep = 0;
            updateScreen = true;
            break;
        }
      }
      break;

    case 6:
      showEditPwrBtm();
      if (value > last) { //encoder plus
        last = value;
        nex = true;
        if (nex && ( millis() - ms_button) > 250) {
          ms_button =  millis();
          beep_(80, 1046);
          nex = false;
          profile.pwr_BOTTOM++;
          if (profile.pwr_BOTTOM >= 100) {
            profile.pwr_BOTTOM = 100;
          }
        }
      }

      if (value < last) { //encoder minus
        last = value;
        prev = true;
        if (prev && ( millis() - ms_button) > 250) {
          ms_button =  millis();
          beep_(80, 1046);
          prev = false;
          profile.pwr_BOTTOM--;
          if (profile.pwr_BOTTOM <= 1) {
            profile.pwr_BOTTOM = 1;
          }
        }
      }

      if (b != modedEncoder::Open) { //encoder button
        switch (b) {
          case modedEncoder::Clicked:
            beep_(80, 1046);
            saveProfilePowerBtm();
            screen = 17;
            updateScreen = true;
            break;
        }
      }
      break;

    case 7:
      showEditPwrTop();
      if (value > last) { //encoder plus
        last = value;
        nex = true;
        if (nex && ( millis() - ms_button) > 250) {
          ms_button =  millis();
          beep_(80, 1046);
          nex = false;
          profile.pwr_TOP++;
          if (profile.pwr_TOP >= 100) {
            profile.pwr_TOP = 100;
          }
        }
      }

      if (value < last) { //encoder minus
        last = value;
        prev = true;
        if (prev && ( millis() - ms_button) > 250) {
          ms_button =  millis();
          beep_(80, 1046);
          prev = false;
          profile.pwr_TOP--;
          if (profile.pwr_TOP <= 1) {
            profile.pwr_TOP = 1;
          }
        }
      }

      if (b != modedEncoder::Open) { //encoder button
        switch (b) {
          case modedEncoder::Clicked:
            beep_(80, 1046);
            saveProfilePowerTop();
            loadProfile();
            screen = 17;
            updateScreen = true;
            break;
        }
      }
      break;

    case 8:
      showPIDMenuHeader();
      if (value > last) { //encoder plus
        last = value;
        nex = true;
        clearLine2();
        if (nex && ( millis() - ms_button) > 250) {
          ms_button =  millis();
          beep_(80, 1046);
          nex = false;
          pidCursor++;
          if (pidCursor > 3) {
            pidCursor = 0;
          }
        }
      }
      if (value < last) { //encoder minus
        last = value;
        prev = true;
        clearLine2();
        if (prev && ( millis() - ms_button) > 250) {
          ms_button =  millis();
          beep_(80, 1046);
          prev = false;
          pidCursor--;
          if (pidCursor < 0) {
            pidCursor = 3;
          }
        }
      }
      showPIDMenu();
      if (b != modedEncoder::Open) { //encoder button
        switch (b) {
          case modedEncoder::Clicked:
            beep_(80, 1046);
            updateScreen = true;
            pidCursorD = 0;
            switch (pidCursor) {
              case 0: screen = 9; pidSelect = 0; break;
              case 1: screen = 9; pidSelect = 1; break;
              case 2: screen = 15; break;
              case 3: screen = 0; break;
            }
            break;
        }
      }
      break;

    case 9:
      showPIDMenuHeader();
      if (value > last) { //encoder plus
        last = value;
        nex = true;
        clearLine2();
        if (prev && ( millis() - ms_button) > 250) {
          ms_button =  millis();
          beep_(80, 1046);
          nex = false;
          switch (pidSelect) {
            case 0:
              pidCursorD++;
              if (pidCursorD > 3) {
                pidCursorD = 0;
              }
              break;

            case 1:
              pidCursorU++;
              if (pidCursorU > 3) {
                pidCursorU = 0;
              }
              break;
          }
        }
      }

      if (value < last) { //encoder minus
        last = value;
        prev = true;
        clearLine2();
        if (prev && ( millis() - ms_button) > 250) {
          ms_button =  millis();
          beep_(80, 1046);
          prev = false;
          switch (pidSelect) {
            case 0:
              pidCursorD--;
              if (pidCursorD < 0) {
                pidCursorD = 3;
              }
              break;

            case 1:
              pidCursorU--;
              if (pidCursorU < 0) {
                pidCursorU = 3;
              }
              break;
          }
        }
      }

      if (b != modedEncoder::Open) { //encoder button
        switch (b) {
          case modedEncoder::Clicked:
            beep_(80, 1046);
            updateScreen = true;
            if (pidSelect == 0) {
              switch (pidCursorD) {
                case 0: screen = 10; pidEdit = 0; break;
                case 1: screen = 10; pidEdit = 1; break;
                case 2: screen = 10; pidEdit = 2; break;
                case 3: screen = 8; break;
              }
              break;
            }
            if (pidSelect == 1) {
              switch (pidCursorU) {
                case 0: screen = 10; pidEdit = 3; break;
                case 1: screen = 10; pidEdit = 4; break;
                case 2: screen = 10; pidEdit = 5; break;
                case 3: screen = 8; break;
              }
            }
        }
      }
      if (pidSelect == 0) {
        showPIDCurD();
      } else if (pidSelect == 1) {
        showPIDCurU();
      }
      break;

    case 10:
      pidEditHeader();
      if (value > last) { //encoder plus
        last = value;
        beep_(40, 1046);
        switch (pidEdit) {
          case 0:
            KpD += increment * 0.01;
            if (KpD >= 255.01) {
              KpD = 255;
            }
            break;
          case 1:
            KiD += increment * 0.01;
            if (KiD >= 255.01) {
              KiD = 255;
            }
            break;
          case 2:
            KdD += increment * 0.01;
            if (KdD >= 255.01) {
              KdD = 255;
            }
            break;
          case 3:
            KpU += increment * 0.01;
            if (KpU >= 255.01) {
              KpU = 255;
            }
            break;
          case 4:
            KiU += increment * 0.01;
            if (KiU >= 255.01) {
              KiU = 255;
            }
            break;
          case 5:
            KdU += increment * 0.01;
            if (KdU >= 255.01) {
              KdU = 255;
            }
            break;
        }
      }

      if (value < last) { //encoder minus
        last = value;
        beep_(40, 1046);
        switch (pidEdit) {
          case 0:
            KpD -= decrement * 0.01;
            if (KpD <= 0.00) {
              KpD = 0.00;
            }
            break;
          case 1:
            KiD -= decrement * 0.01;
            if (KiD <= 0.00) {
              KiD = 0.00;
            }
            break;
          case 2:
            KdD -= decrement * 0.01;
            if (KdD <= 0.00) {
              KdD = 0.00;
            }
            break;
          case 3:
            KpU -= decrement * 0.01;
            if (KpU <= 0.00) {
              KpU = 0.00;
            }
            break;
          case 4:
            KiU -= decrement * 0.01;
            if (KiU <= 0.00) {
              KiU = 0.00;
            }
            break;
          case 5:
            KdU -= decrement * 0.01;
            if (KdU <= 0.00) {
              KdU = 0.00;
            }
            break;
        }
      }
      showPIDEdit();
      if (b != modedEncoder::Open) { //encoder button
        switch (b) {
          case modedEncoder::Clicked:
            beep_(80, 1046);
            switch (pidEdit) {
              case 0: savePidPBtm(); break;
              case 1: savePidIBtm(); break;
              case 2: savePidDBtm(); break;
              case 3: savePidPTop(); break;
              case 4: savePidITop(); break;
              case 5: savePidDTop(); break;
            }
            screen = 9;
            updateScreen = true;
            break;
        }
      }
      break;

    case 11:
      rampScr();
      if (flag) {
        if (tc1 > profile.temperatureStep[currentStep - 1]
            && currentStep != 1) {
          startTemp = profile.temperatureStep[currentStep - 2];
        } else {
          startTemp = tc1;
        }
        counter = 0;
        analogWrite(4, 128);
        flag = false;
      }

      if (!flagB) {
        setpointRampB = 0;
        counterB = 0;
        startTempB = tc2;
        flagB = true;
      }

      if (profile.Setpoint2 > setpointRampB
          && (currentMillis - previousMillisB) > 1000 / (profile.rampRateStep[0] * 0.1)) {
        previousMillisB = currentMillis;
        counterB = counterB + 1;
        setpointRampB = counterB + startTempB;
        Setpoint22 = setpointRampB;
        showRampTemp(setpointRampB, Setpoint22, 1);
        showRampTemp(tc1, tc1, 0);
      }

      if (!TopStart
          && tc2 >= profile.Setpoint2) {
        Setpoint22 = profile.Setpoint2;
        showRampTemp(Setpoint22, Setpoint22, 1);
        showRampTemp(tc1, tc1, 0);
        beep_(40, 1046);
        flag = true;
        TopStart = true;
      }

      if (TopStart
          && (currentMillis - previousMillis) > 1000 / (profile.rampRateStep[currentStep - 1] * 0.1)) {
        previousMillis = currentMillis;

        if (setpointRamp < profile.temperatureStep[currentStep - 1]) {
          counter = counter + 1;
          setpointRamp = counter + startTemp;
        }
        else if (setpointRamp >= profile.temperatureStep[currentStep - 1]) {
          setpointRamp = profile.temperatureStep[currentStep - 1];
        }
        Setpoint1 = setpointRamp;
        showRampTemp(setpointRamp, Setpoint1, 0);
      }

      if (setpointRamp >= profile.temperatureStep[currentStep - 1]
          && (TopStart)) {
        screen = 12;
      }

      if (profile.temperatureStep[currentStep - 1] >= 180) {
        analogWrite(4, 85);
      }

      if (!toggleSwitchState) {
        beep_(200, 1094);
        delay(100);
        beep_(40, 1046);
        updateScreen = true;
        P_D = 0;
        P_U = 0;
        previousMillis = 0;
        previousMillisB = 0;
        screen = 0;
        Status = 0;
      }
      break;

    case 12:
      Setpoint1 = profile.temperatureStep[currentStep - 1];
      if ((Input1 >= profile.temperatureStep[currentStep - 1])) {
        counter = 0;
        screen = 13;
      }

      if (!toggleSwitchState) {
        beep_(200, 1094);
        delay(100);
        beep_(40, 1046);
        updateScreen = true;
        P_D = 0;
        P_U = 0;
        previousMillis = 0;
        previousMillisB = 0;
        screen = 0;
        Status = 0;
      }
      break;

    case 13:
      if (currentMillis - previousMillis > 1000) {
        previousMillis = currentMillis;
        counter = counter + 1;
        int8_t d = profile.dwellTimerStep[currentStep - 1] - counter;
        showDwellCount(d);
      }
      if (counter == profile.dwellTimerStep[currentStep - 1]) {
        counter = 0;
        setpointRamp = 0;
        if (profile.profileSteps == currentStep) {
          currentStep = 1;
          flag = false;
          TopStart = false;
          flagB = false;
          screen = 14;
          nextRead3 = millis();
          PORTC &= B11111101;             //protection Relay LOW
          updateScreen = true;
        } else {
          currentStep++;
          beep_(100, 1046);
          flag = true;
          screen = 11;
          updateScreen = true;
        }
      }

      if (!toggleSwitchState) {
        beep_(200, 1094);
        delay(100);
        beep_(40, 1046);
        updateScreen = true;
        P_D = 0;
        P_U = 0;
        previousMillis = 0;
        previousMillisB = 0;
        screen = 0;
        Status = 0;
      }
      break;

    case 14:
      Status = 0;
      P_D = 0;
      P_U = 0;
      previousMillis = 0;
      previousMillisB = 0;

      showReflowComplete();

      //      if (millis() > nextRead3) {
      //        nextRead3 = millis() + SENSOR_SAMPLING_TIME;
      lcdTemp();
      //      }

      if (!toggleSwitchState) {
        beep_(200, 1094);
        delay(100);
        beep_(40, 1046);
        updateScreen = true;
        screen = 0;
      }
      break;

    case 15:
      if (Status == 3) {
        showResetHeader();
      }

      if (Status != 3) {
        showPIDTuneHeader();
        if (value > last) { //encoder plus
          last = value;
          nex = true;
          clearLine2();
          if (nex && ( millis() - ms_button) > 250) {
            ms_button =  millis();
            beep_(80, 1046);
            nex = false;
            selection++;
            if (selection > 2) {
              selection = 0;
            }
          }
        }

        if (value < last) { //encoder minus
          last = value;
          prev = true;
          clearLine2();
          if (prev && ( millis() - ms_button) > 250) {
            ms_button =  millis();
            beep_(80, 1046);
            prev = false;
            selection--;
            if (selection < 0) {
              selection = 2;
            }
          }
        }
        showPIDTuneSel();
      }

      if (b != modedEncoder::Open) { //encoder button
        switch (b) {
          case modedEncoder::Clicked:
            if (Status != 3) {
              beep_(40, 1046);
              updateScreen = true;
              if (selection == 0) {
                Status = 2;
                screen = 16;
                PORTC |= B00000010;                    //protection Relay HIGH
                nextRead3 = millis();
                nextRead2 = millis();
                tuning2 = false;
                tuner2.setParameters(
                  0,
                  75,
                  75,
                  SENSOR_SAMPLING_TIME,
                  SENSOR_SAMPLING_TIME,
                  SENSOR_SAMPLING_TIME / 10,
                  SENSOR_SAMPLING_TIME);
                regulator2.setDirection(0);
                regulator2.setDt(SENSOR_SAMPLING_TIME);
                regulator2.setLimits(0, 150);
                startSerialCom();
              } else if (selection == 1) {
                analogWrite(4, 128);
                Status = 4;
                screen = 16;
                PORTC |= B00000010;                   //protection Relay HIGH
                nextRead3 = millis();
                nextRead2 = millis();
                tuning1 = false;
                tuner1.setParameters(
                  0,
                  75,
                  75,
                  SENSOR_SAMPLING_TIME,
                  SENSOR_SAMPLING_TIME,
                  SENSOR_SAMPLING_TIME / 10,
                  SENSOR_SAMPLING_TIME);
                regulator1.setDirection(0);
                regulator1.setDt(SENSOR_SAMPLING_TIME);
                regulator1.setLimits(0, 150);
                startSerialCom();
              } else if (selection == 2) {
                screen = 8;
              }
            }
            break;
          case modedEncoder::DoubleClicked:
            if (Status == 3) {
              beep_(80, 1046);
              format = true;
              updateScreen = true;
            }
            break;
          case modedEncoder::Released:
            if (Status == 3) {
              Status = 0;
              beep_(120, 1046);
              updateScreen = true;
              screen = 0;
            }
            break;
        }
      }
      break;

    case 16:
      showTuningHeader();
      if (b != modedEncoder::Open) { //encoder button
        switch (b) {
          case modedEncoder::Released:
            beep_(120, 1046);
            updateScreen = true;
            delay(200);
            screen = 0;
            Status = 0;
            selection = 0;
            break;
        }
      }
      break;

    case 17:
      profileEditSelScr();
      if (value > last) { //encoder plus
        last = value;
        nex = true;
        clearLine2();
        if (nex && ( millis() - ms_button) > 250) {
          ms_button =  millis();
          beep_(80, 1046);
          nex = false;
          selection2++;
          if (selection2 > 7) {
            selection2 = 0;
          }
        }
      }

      if (value < last) { //encoder minus
        last = value;
        prev = true;
        clearLine2();
        if (prev && ( millis() - ms_button) > 250) {
          ms_button =  millis();
          beep_(80, 1046);
          prev = false;
          selection2--;
          if (selection2 <= 0) {
            selection2 = 7;
          }
        }
      }

      profileEditSel();

      if (b != modedEncoder::Open) { //encoder button
        switch (b) {
          case modedEncoder::Clicked:
            beep_(80, 1046);
            updateScreen = true;
            switch (selection2) {
              case 0: screen = 1; break;
              case 1: screen = 2; break;
              case 2: screen = 3; editStep = 0; break;
              case 3: screen = 4; editStep = 0; break;
              case 4: screen = 5; editStep = 0; break;
              case 5: screen = 6; break;
              case 6: screen = 7; break;
              case 7: screen = 0; break;
            }
            break;
        }
      }
      break;
  }

  if (Status == 0) {
    if (Input1 > 50) {
      if (((PIND & B00010000) >> 4) == LOW) {
        PORTD |= B00010000;
      }
    }
    else if (Input1 < 50) {
      if (((PIND & B00010000) >> 4) == HIGH) {
        PORTD &= B11101111;
      }
    }

    if (((PINC & B00000010) >> 1) == HIGH) {
      PORTC &= B11111101;                                   //protection Relay LOW
    }
  } else if (Status == 1) {

    if (millis() > nextRead2) {
      nextRead2 = millis() + GRAPHICS_SAMPLING_TIME;
      if (curCount == 0) {
        if (TopStart == true) {
          showTopIcon(true);
        }
        if (flagB == true) {
          showBottomIcon(true);
        }
        curCount = 1;
      } else {
        showTopIcon(false);
        showBottomIcon(false);
        curCount = 0;
      }
    }

    showPulseOut(P_U_Serial, 0);
    showPulseOut(P_D_Serial, 1);

    int x = Setpoint1 - Input1;
    int gap1 = abs(x);
    if (gap1 < 10) {
      regulator1.Kp = KpU;
      regulator1.Ki += KiU;
      regulator1.Kd = KdU;
    } else {
      regulator1.Kp = TUNE_P_PID;
      regulator1.Ki += TUNE_I_PID;
      regulator1.Kd = TUNE_D_PID;
    }

    int y = Setpoint22 - Input2;
    int gap2 = abs(y);
    if (gap2 < 2) {
      regulator2.Kp = KpD;
      regulator2.Ki += KiD;
      regulator2.Kd = KdD;
    } else {
      regulator2.Kp = TUNE_P_PID;
      regulator2.Ki += TUNE_I_PID;
      regulator2.Kd = TUNE_D_PID;
    }

    lcdTemp();

    if (millis() > nextRead3) {
      nextRead3 = millis() + SENSOR_SAMPLING_TIME;

      regulator1.setpoint = Setpoint1;
      regulator1.input = Input1;

      regulator2.setpoint = Setpoint22;
      regulator2.input = Input2;

      if (flagB) {
        P_D = regulator2.getResult();
        P_D_Serial = map(P_D, 0, 150 , 0 , 100);
        BottomHeaterON(P_D);
      }

      if (TopStart) {
        P_U = regulator1.getResult();
        P_U_Serial = map(P_U, 0, 150 , 0 , 100);
        TopHeaterON(P_U);
      }
      Serial.println(buf);
    }

    if (!flagB) {
      P_D = 0;
      BottomHeaterOFF();
    }

    if (!TopStart) {
      P_U = 0;
      TopHeaterOFF();
    }
  } else if (Status == 2) {

    lcdTemp();
    if (millis() > nextRead3) {
      nextRead3 = millis() + SENSOR_SAMPLING_TIME;
      if (!tuning2) {
        regulator2.Kp = TUNE_P_PID;
        regulator2.Ki += TUNE_I_PID;
        regulator2.Kd = TUNE_D_PID;
        regulator2.setpoint = tuneSP;
        regulator2.input = Input2;
        P_D = regulator2.getResult();
      } else if (tuning2) {
        tuner2.setInput(Input2);
        tuner2.compute();
        P_D = tuner2.getOutput();
      }
      BottomHeaterON(P_D);
      Serial.println(buf);
    }

    if (millis() > nextRead2) {
      nextRead2 = millis() + GRAPHICS_SAMPLING_TIME;
      if (curCount == 0) {
        showBottomIcon(true);
        curCount = 1;
      } else {
        showBottomIcon(false);
        curCount = 0;
      }
    }

    double tuneGap2 = abs(tuneSP - Input2);
    if (tuneGap2 < 0.5) {
      tuning2 = true;
      tuningTrueDisp();
    }
    if (tuning2) {
      tunerAccu2 = tuner2.getAccuracy();
      if (tuner2.getAccuracy() > 97) {
        KpD = tuner2.getPID_p();
        KiD = tuner2.getPID_i();
        KdD = tuner2.getPID_d();
        tuner2.reset();
        savePidPBtm();
        savePidIBtm();
        savePidDBtm();
        loadPIDParam();
        tuning2 = false;
        screen = 0;
        Status = 0;
        selection = 0;
        BottomHeaterOFF();
        updateScreen = true;
      }
    }
    tuneScreenShow(P_D, tunerAccu2, 1);
    TopHeaterOFF();
  } else if (Status == 3) {

    if (millis() > nextRead2) {
      nextRead2 = millis() + GRAPHICS_SAMPLING_TIME;
      lcd.setCursor(2, 1);
      if (curCount == 0) {
        curCount = 1;
        lcd.print(F("d-clk> cont."));
      } else {
        lcd.print(F("l-prs> cancl"));
        curCount = 0;
      }
    }

    if (format) {
      EEPROM.write(eeprom_id_address, eeprom_reset);
      while (1) {
        initEeprom();
        if (initEepromComplete) {
          initEepromComplete = false;
          rstScreen();
          sReset();
        }
      }
    }
  }

  else if (Status == 4) {

    lcdTemp();
    if (millis() > nextRead3) {
      nextRead3 = millis() + SENSOR_SAMPLING_TIME;
      if (tuning1 == false) {
        regulator1.Kp = TUNE_P_PID;
        regulator1.Ki += TUNE_I_PID;
        regulator1.Kd = TUNE_D_PID;
        regulator1.setpoint = tuneSP;
        regulator1.input = Input1;
        P_U = regulator1.getResult();
      } else {
        tuner1.setInput(Input1);
        tuner1.compute();
        P_U = tuner1.getOutput();
      }
      TopHeaterON(P_U);
      Serial.println(buf);
    }
    if (millis() > nextRead2) {
      nextRead2 = millis() + GRAPHICS_SAMPLING_TIME;
      if (curCount == 0) {
        showTopIcon(true);
        curCount = 1;
      } else {
        showTopIcon(false);
        curCount = 0;
      }
    }

    double tuneGap1 = abs(tuneSP - Input1);
    if (tuneGap1 < 0.5) {
      tuning1 = true;
      tuningTrueDisp();
    }

    if (tuning1) {
      tunerAccu1 = tuner1.getAccuracy();
      if (tuner1.getAccuracy() > 98) {
        KpU = tuner1.getPID_p();
        KiU = tuner1.getPID_i();
        KdU = tuner1.getPID_d();
        tuner1.reset();
        savePidPTop();
        savePidITop();
        savePidDTop();
        loadPIDParam();
        screen = 0;
        Status = 0;
        selection = 0;
        TopHeaterOFF();
        tuning1 = false;
        updateScreen = true;
      }
    }
    tuneScreenShow(P_U, tunerAccu1, 0);
    BottomHeaterOFF();
  }
  goto labelbegin;
}


#pragma GCC pop_options
